#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/Xaw/Simple.h>

#include "dftype.h"
#include "dfxfunc.h"

void Buf_Init(DfBuf *p)
{
  memset(p, 0, sizeof(DfBuf));
}


void ListClear(void *ptr)
{
  DfList *buf;

  buf = ptr;
  buf->next = ptr;
  buf->prev = ptr;
}

void *ListAdd(void *a, void *b)
{
  DfList *ptr;
  DfList *follow;

  if(a == b){
    return a;
  }
  ptr = a;
  follow = b;

  follow->next = ptr->next;
  follow->prev = ptr;
  follow->next->prev = follow;
  ptr->next = follow;

  return follow;
}

void *ListDel(void *a)
{
  DfList *ptr;
  DfList *prev;
  DfList *next;

  ptr = a;
  prev = ptr->prev;
  next = ptr->next;

  prev->next = next;
  next->prev = prev;

  return ptr;
}

void *NextBuffer(void *ptr)
{
  DfList *buf;

  buf = ptr;
  return buf->next;
}

void *PrevBuffer(void *ptr)
{
  DfList *buf;

  buf = ptr;
  return buf->prev;
}

int RectIsOverlap(XRectangle *a, XRectangle *b)
{
  int a_w;
  int b_w;
  int a_h;
  int b_h;

  a_w = a->x + a->width;
  b_w = b->x + b->width;
  a_h = a->y + a->height;
  b_h = b->y + b->height;

  if(a->x < b_w && b->x < a_w && a->y < b_h && b->y < a_h){
    return 1;
  }
  return 0;
}

void RectAdd(XRectangle *a, XRectangle *b)
{
  int x, y, cx, cy;
  int a_w;
  int b_w;
  int a_h;
  int b_h;

  a_w = a->x + a->width;
  b_w = b->x + b->width;
  a_h = a->y + a->height;
  b_h = b->y + b->height;

  if(a->x < b->x){
    x = a->x;
  }else{
    x = b->x;
  }

  if(a->y < b->y){
    y = a->y;
  }else{
    y = b->y;
  }

  cx = a_w;
  if(a_w < b_w){
    cx = b_w;
  }
  cy = a_h;
  if(a_h < b_h){
    cy = b_h;
  }

  a->x = x;
  a->y = y;
  a->width = cx - x;
  a->height = cy - y;
}

static int CheckEdge(short int x, unsigned short int x_len, short int y, unsigned short int y_len)
{
  int s;
  int n;
  int m;

  n = x + x_len;
  m = y + y_len;
  
  s = 0;
  if(y <= x && x <= m){
    s |= 1;
  }
  if(y <= n && n <= m){
    s |= 2;
  }
  return s;
}

void RectSub(XRectangle *a, XRectangle *b)
{
  int i;
  int j;
  int cx;
  int cy;

  i = CheckEdge(a->x, a->width, b->x, b->width);
  j = CheckEdge(a->y, a->height, b->y, b->height);
  
  switch(i<<4 | j){
  case 0x13:
    cx = a->x + a->width;
    a->x = b->x + b->width;
    a->width = cx - a->x;
    break;
  case 0x23:
    a->width = b->x - a->x;
    break;
  case 0x31:
    cy = a->y + a->height;
    a->y = b->y + b->height;
    a->height = cy - a->y;
    break;
  case 0x32:
    a->height = b->y - a->y;
    break;
  case 0x33:
    a->width = 0;
    a->height = 0;
    break;
  }
}


