#include <malloc.h>
#include <stdio.h>

#include <X11/Intrinsic.h>

#include "dftype.h"
#include "dfxfunc.h"

#include <errno.h>

typedef struct _tmpList{
  DfFile *files;
  DfStr s;
  int num;
  int alloced;
}tmplist;

static int SetupListing(DfFileList *pList, tmplist *tmp);
static int AddList(tmplist *tmp, const char *filename, struct stat *st);
static int FinishListing(DfFileList *pList, tmplist *tmp);

void DfxFree(DfFileList *list)
{
  bFree(list->files);
  bFree(list->pszFilenames);
  bFree(list->pszCurDir);
  bFree(list);
}

int GetFiles(DfFileList *list, const char *new_dir)
{
  DfStr s;
  DIR *dir;
  struct dirent *d;
  struct stat st;
  tmplist t;
  int idx;
  df_stat ds;
  DfFile *files;

  files = list->files;

  ds = list->stat;
  list->stat = DS_WAIT_DIR;
  /* reuse memory */
  if(list->pszCurDir){
    Str_InitPsz(&s, list->pszCurDir);
  }else{
    Str_Init(&s);
  }

  SetLastSlash(&s);
  idx = Str_Length(&s);/* for restore */
  if(new_dir && new_dir != list->pszCurDir){
    if(IsAbsPath(new_dir)){
      Str_Overwrite(&s, 0, new_dir);
    }else{
      Str_Add(&s, new_dir);
    }
  }

  Str_RegPath(&s);

  dir = opendir(Str_Get(&s));
  if(!dir){
    fprintf(stderr, "oendir(%s) to fail. %s\n", Str_Get(&s), strerror(errno));
    /* restore cwd  */
    Str_SetLength(&s, idx);
    Str_Shrink(&s);
    return 0;
  }

  SetLastSlash(&s);
  idx = Str_Length(&s);

  SetupListing(list, &t);

  while((d = readdir(dir)) != 0){
    Str_Overwrite(&s, idx, d->d_name);
    stat(Str_Get(&s), &st);
    if(!AddList(&t, d->d_name, &st)){
      break;
    }
  }
  closedir(dir);
  FinishListing(list, &t);
  bFree(files);
  Str_SetLength(&s, idx);
  Str_Shrink(&s);
  list->pszCurDir = Str_Get(&s);

  list->stat = ds;
  /*  it's temporary cursor fix. */
  if(list->nFiles <= list->nCur){
    list->nStart = 0;
    list->nCur = list->nFiles -1;
  }

  return 1;
}

static int AddList(tmplist *tmp, const char *filename, struct stat *st)
{
  DfFile *files;
  int n;
  int c;

  /* extend list */
  n = tmp->num;
  files = tmp->files;
  if(tmp->alloced <= n){
    files = bReAlloc(files, (sizeof(*files) * (tmp->alloced * 2)));
    if(!files){
      return 0;
    }
    tmp->alloced *= 2;
  }

  c = COLOR_FILE;
  switch(st->st_mode & S_IFMT){
  case S_IFREG:
    c = COLOR_FILE;
    if(st->st_mode & 0111){
      c = COLOR_X;
    }
    break;
  case S_IFDIR:
    c = COLOR_DIR;
    break;
  case S_IFLNK:
    c = COLOR_LINK;
    break;
  case S_IFBLK:
  case S_IFCHR:
    c = COLOR_DEV;
    break;
  }

  files[n].pszName = NULL;
  files[n].attr = st->st_mode;
  files[n].color = colors[c];
  files[n].size = st->st_size;
  files[n].owner = st->st_uid;
  files[n].group = st->st_gid;
  files[n].atime = st->st_atime;
  files[n].mtime = st->st_mtime;
  files[n].ctime = st->st_ctime;
  files[n].nSelect = 0;
  files[n].pszName = (char *)(Str_Length(&tmp->s));/* hold index */

  Str_Add(&tmp->s, filename);
  Str_AddChar(&tmp->s, '\0');

  tmp->files = files;

  tmp->num++;

  return 1;
}

static int SetupListing(DfFileList *list, tmplist *tmp){
  Str_Init(&tmp->s);

  tmp->files = bMalloc(sizeof(DfFile) * 32);
  tmp->alloced = 32;
  tmp->num = 0;

  return 0;
}

static int FinishListing(DfFileList *list, tmplist *tmp)
{
  int n;
  int max;
  DfFile *f;
  void *p;

  list->nFiles = tmp->num;

  p = bReAlloc(tmp->files, sizeof(DfFile) * tmp->num);
  list->files = p ? p : tmp->files;

  Str_Shrink(&tmp->s);
  bFree(list->pszFilenames);
  list->pszFilenames = Str_Get(&tmp->s);

  /*
   * Adjust index
   */
  max = list->nFiles;

  f = list->files;
  for(n = 0; n < max; n++){
    f->pszName += (int)list->pszFilenames;
    f++;
  }

  return 1;
}
