#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/keysym.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Simple.h>

#include "dftype.h"
#include "dfxfunc.h"

Widget toplevel;
XFontSet vFontSet;
int vnFontHeight;
int vnFontBase;
int colors[COLOR_MAX];
DfWindow *dfx;
int mod_keys;

extern void action_quit(Widget w, XEvent *ev, String *s, Cardinal *c);
static void AppInitialize(Widget w);
static void wndResize(Widget w, XtPointer p, XEvent *event, Boolean *cont);
static void wndKeyPressProc(Widget w, XtPointer p, XEvent *event, Boolean *cont);
static void wndKeyReleaseProc(Widget w, XtPointer p, XEvent *event, Boolean *cont);
static void wndExpose(Widget w, XtPointer p, XEvent *event, Boolean *cont);
static void wndDrawProc(DfWindow *dfx, Display *display, Widget w, GC gc);

static char *fallback_resources[] ={
//  "dfx.width: 320",
//  "dfx.hieght: 200",
  "dfx.dlg.dialog.width: 160",
  "dfx.dlg.dialog.height: 200",
  "dfx.dlg.dialog.horizDistance: 2",
  "dfx.dlg.dialog.vertDistance: 2",
  "*text.colums: 24",
  "*text.value: \".\"",
  NULL
};

void test(void)
{
  XRectangle a[]={
    {100, 100, 100, 100},
    {-1, -1, 0, 0}
  };
  XRectangle b[]={
    {  0,   0, 300, 100},
    {  0,   0, 300, 300},
    {  0,   0, 300, 150},
    {  0, 150, 300, 100},
    {  0,   0, 150, 300},
    {150,   0,  50, 300},
    {  0,   0, 150, 150},
    {-1, -1, 0, 0}
  };
  XRectangle x, y;
  int hit;
  int i, j;
  for(i = 0; ; i++){
    y = b[i];
    if(y.width ==0 && y.height == 0){
      break;
    }
    for(j = 0; ; j++){
      x = a[j];
      if(x.width ==0 && x.height == 0){
	break;
      }
      hit = RectIsOverlap(&x, &y);
      printf("Hit : (%d, %d, %d, %d), (%d, %d, %d, %d) %s\n",
	     x.x, x.y, x.width, x.height,
	     y.x, y.y, y.width, y.height,
	     hit ? "hit" : "no");

      x = a[j];
      RectAdd(&x, &y);
      printf("Add: (%d, %d, %d, %d)\n",
	     x.x, x.y, x.width, x.height);

      x = a[j];
      RectSub(&x, &y);
      printf("Sub: (%d, %d, %d, %d)\n",
	     x.x, x.y, x.width, x.height);
    }
  }

}

int main(int argc, char **argv)
{
  XtAppContext ac;

  test();
  toplevel = XtAppInitialize(&ac, "dfx", NULL, 0,
			     &argc, argv, fallback_resources, NULL, 0);
  AppInitialize(toplevel);
  Wnd_Create(toplevel, 150, 120);
  //  XtAppAddActions(ac, actions, XtNumber(actions));
  //  XtOverrideTranslations(toplevel, XtParseTranslationTable(translate));
  XtRealizeWidget(toplevel);
  XtAppMainLoop(ac);

  return 0;
}

static void AppInitialize(Widget w)
{
  char **miss;
  int miss_charset;
  char *def;
  XFontSetExtents *fse;
  Display *d;

  d = XtDisplay(w);
  colors[COLOR_BK] = xAllocColor(d, "white");
  colors[COLOR_FILE] = xAllocColor(d, "black");
  colors[COLOR_RO] = xAllocColor(d, "red");
  colors[COLOR_X] = xAllocColor(d, "green");
  colors[COLOR_DIR] = xAllocColor(d, "blue");
  colors[COLOR_CURBG] = xAllocColor(d, "black");
  colors[COLOR_CURFG] = xAllocColor(d, "white");
  colors[COLOR_SELBG] = xAllocColor(d, "blue");
  colors[COLOR_SELFG] = xAllocColor(d, "white");
  colors[COLOR_MARK] =  xAllocColor(d, "yellow");

  vFontSet = XCreateFontSet(
			    d,
			    "-*-*-medium-r-normal--12-*",
			    &miss, &miss_charset, &def);

  fse = XExtentsOfFontSet(vFontSet);
  vnFontHeight = fse->max_logical_extent.height;  
  vnFontBase = -fse->max_logical_extent.y;
}

/* will be use XtGrabKeyboard() */
static void wndKeyPressProc(Widget w, XtPointer p, XEvent *event, Boolean *cont)
{
  DfBuf *buf;
  KeySym k;

  k = XLookupKeysym((XKeyEvent*)event, 0);

  switch(k){
  case XK_Shift_L:
  case XK_Shift_R:
    mod_keys |= DFMOD_SHIFT;
    break;

  case XK_Control_L:
  case XK_Control_R:
    mod_keys |= DFMOD_CTRL;
    break;

  case XK_Alt_L:
  case XK_Alt_R:
    mod_keys |= DFMOD_ALT;
    break;
  }

  buf = dfx->active;
  buf->v->keybd_event(buf, w, k, cont);
}

static void wndKeyReleaseProc(Widget w, XtPointer p, XEvent *event, Boolean *cont)
{
  KeySym k;

  k = XLookupKeysym((XKeyEvent*)event, 0);

  switch(k){
  case XK_Shift_L:
  case XK_Shift_R:
    mod_keys &= ~DFMOD_SHIFT;
    break;

  case XK_Control_L:
  case XK_Control_R:
    mod_keys &= ~DFMOD_CTRL;
    break;

  case XK_Alt_L:
  case XK_Alt_R:
    mod_keys &= ~DFMOD_ALT;
    break;
  }
}

static void wndExpose(Widget w, XtPointer p, XEvent *event, Boolean *cont)
{
  XExposeEvent *ev;
  GC gc;

  ev = (XExposeEvent*)event;

  if(ev->count){
    /* add to clip mask and exit. */
    if(dfx->fClipping){
      AddRect(&dfx->clip, ev->x, ev->y, ev->width, ev->height);
    }else{
      dfx->fClipping = 1;
      dfx->clip.x = ev->x;
      dfx->clip.y = ev->y;
      dfx->clip.width = ev->width;
      dfx->clip.height = ev->height;
    }
    return;
  }

  /* drawing */
  dfx->clip.x = ev->x;
  dfx->clip.y = ev->y;
  dfx->clip.width = ev->width;
  dfx->clip.height = ev->height;

  gc = DefaultGC(ev->display, DefaultScreen(ev->display));

  wndDrawProc(dfx, ev->display, w, gc);
}

static void wndDrawProc(DfWindow *dfx, Display *display, Widget w, GC gc)
{
  DfBuf *start;
  DfBuf *buf;

  XSetClipRectangles(display, gc, 0, 0, &dfx->clip, 1, Unsorted);

  start = buf = dfx->lists;
  do{
    if(buf->visible){
      buf->v->draw(buf, display, w, gc);
    }
    buf = NextBuffer(buf);
  }while(start != buf);

  dfx->fClipping = 0;
}

static void wndResize(Widget w, XtPointer p, XEvent *event, Boolean *cont){
  XConfigureEvent *ev;
  DfFileList *list;
  int cur;
  int n;

  switch(event->type){
  case ConfigureNotify:
    ev = (XConfigureEvent*)event;

    dfx->width = ev->width;
    dfx->height = ev->height;

    /* temporary */
    list = (DfFileList*)dfx->lists;
    list->b.rc.width = dfx->width;
    list->b.rc.height = dfx->height;

    cur = list->nCur;
    if(list->b.rc.height < vnFontHeight){
      list->nStart = cur;
    }else{
      n = list->b.rc.height / vnFontHeight;
      if(n){
        n--;
      }
      if(list->nStart + n < cur){
        if(cur - n <= 0){
          n = 0;
        }else{
          n = cur - n;
        }
        list->nStart = n;
      }
    }
    break;
  }
}


/* filer window */
void BufSetActive(DfBuf *b)
{
  dfx->active = b;
  b->v->activate(b);
}

void Wnd_Create(Widget parent, int cx, int cy)
{
  Widget w;
  DfFileList *list;
  char buf[256];/* for current directory */

  w = XtVaCreateManagedWidget("main", simpleWidgetClass, parent,
			      XtNwidth, cx,
			      XtNheight, cy,
			      XtNborderWidth, 4,
			      NULL);
  dfx = bMalloc(sizeof(*dfx));
  Wnd_Initialize(w, dfx);

  dfx->width = cx;
  dfx->height = cy;

  list = bMalloc(sizeof(*list));
  dfx->lists = (DfBuf*)list;

  FL_Init(list);
  list->nCur = 0;
  list->nStart = 0;
  list->pszCurDir = NULL;
  list->b.rc.x = 0;
  list->b.rc.y = 0;
  list->b.rc.width = cx;
  list->b.rc.height = cy;
  list->b.visible = 1;

  dfx->nBuf++;

  getcwd(buf, sizeof(buf));
  GetFiles(list, buf);

  BufSetActive((DfBuf*)list);
}

void Wnd_Initialize(Widget w, DfWindow *dfx)
{

  memset(dfx, 0, sizeof(*dfx));

  /* initialize */
  dfx->nBuf = 0;
  dfx->w = w;

  /* add event handler */
  XtAddEventHandler(w, KeyPressMask, False, wndKeyPressProc, NULL);
  XtAddEventHandler(w, KeyReleaseMask, False, wndKeyReleaseProc, NULL);
  XtAddEventHandler(w, ExposureMask, False, wndExpose, NULL);
  XtAddEventHandler(w, StructureNotifyMask, False, wndResize, NULL);
}


