#include <malloc.h>
#include <string.h>
#include <stdio.h>

typedef struct _meminfo{
  void *p;
  int size;
  const char *file;
  const char *func;
  int line;
}MEMINFO;

MEMINFO *vMemInfo;
int vnMems;

static void AddMemInfo(void *p, int size, const char *file, const char *func, int line);
static void DeleteMemInfo(MEMINFO *p);
static MEMINFO *FindMemInfo(void *p);

void *dbgMalloc(int size, const char *file, const char *func, int line)
{
  void *p;
  p = malloc(size);
  AddMemInfo(p, size, file, func, line);
  return p;
}

void *dbgReAlloc(void *ptr, int size, const char *file, const char *func, int line)
{
  void *p;
  MEMINFO *info;

  info = FindMemInfo(ptr);
  if(info == NULL){
    fprintf(stderr, "BUG: realloc() called with invalid pointer. %s %s %d\n",
	    file, func, line);
    return NULL;
  }
    
  p = realloc(ptr, size);
  if(p){
    info->p = p;
    info->size = size;
    info->file = file;
    info->func = func;
    info->line = line;
  }

  return p;
}

void dbgFree(void *ptr, const char *file, const char *func, int line)
{
  MEMINFO *info;

  if(!ptr){
    return;
  }
  info = FindMemInfo(ptr);
  if(info == NULL){
    fprintf(stderr, "BUG: free() called with invalid pointer. %s %s %d\n",
	    file, func, line);
    return;
  }

  free(ptr);
    
  DeleteMemInfo(info);
}

void DumpMemInfo(void)
{
  MEMINFO *p;
  int n;
  int total;

  total = 0;
  p = vMemInfo;
  for(n = 0; n < vnMems; n++){
    fprintf(stderr, "%d: %dBytes %s:%s() %d\n",
	    n, p->size, p->file, p->func, p->line);
    total += p->size;
    p++;
  }
  fprintf(stderr, "Total %d.\n", total);
}

void MemModInfo(void *ptr, const char *file, const char *func, int line)
{
  MEMINFO *info;

  info = FindMemInfo(ptr);
  if(info == NULL){
    fprintf(stderr, "BUG: realloc() called with invalid pointer. %s %s %d\n",
	    file, func, line);
    return;
  }

  info->file = file;
  info->func = func;
  info->line = line;
}

static void AddMemInfo(void *p, int size, const char *file, const char *func, int line)
{
  MEMINFO *newInfo;

  if(vnMems){
    newInfo = realloc(vMemInfo, sizeof(MEMINFO) * (vnMems + 1));
  }else{
    newInfo = malloc(sizeof(MEMINFO));
  }
    
  if(!newInfo){
    fprintf(stderr, "not enough memory at AddMemInfo.\n");
    return;
  }
  vMemInfo = newInfo;
  newInfo = vMemInfo + vnMems;
  newInfo->p = p;
  newInfo->size = size;
  newInfo->file = file;
  newInfo->func = func;
  newInfo->line = line;
  vnMems++;
}

static void DeleteMemInfo(MEMINFO *p)
{
  int n;

  n = p -vMemInfo;
  n = (vnMems - n - 1) * sizeof(MEMINFO);
  memmove(p, p + 1, n);

  vnMems--;
}

static MEMINFO *FindMemInfo(void *ptr)
{
  MEMINFO *p;
  int n;

  p = vMemInfo;
  for(n = 0; n < vnMems; n++){
    if(p->p == ptr){
      return p;
    }
    p++;
  }
  return NULL;
}
