#include <X11/Intrinsic.h>

#include "dftype.h"
#include "dfxfunc.h"


int IsAbsPath(const char *path)
{
  return path[0] == '/';
}

int IsEndSlash(DfStr *path)
{
  int n;

  n = Str_Length(path);

  /*
   * ToDo: support multi byte char
   */
  if(0 < n && Str_Get(path)[n - 1] == '/'){
    return n;
  }
  return 0;
}

void SetLastSlash(DfStr *path)
{
  if(Str_Length(path) != 0 && !IsEndSlash(path)){
    Str_AddChar(path, '/');
  }
  return;
}

void AddRect(XRectangle *rc, int x, int y, int width, int height)
{
  int ox;
  int oy;
  int cx;
  int cy;

  ox = rc->x + rc->width;
  oy = rc->y + rc->height;
  cx = x + width;
  cy = y + height;

  if(x < rc->x){
    rc->x = x;
  }

  if(y < rc->y){
    rc->y = y;
  }
  
  if(ox < cx){
    ox = cx;
  }
  rc->width = ox - rc->x;
  if(oy < cy){
    oy = cy;
  }
  rc->height = oy - rc->y;

}

unsigned long xAllocColor(Display *d, const char *color)
{
  Colormap c_map;
  XColor c1, c2;

  c_map = DefaultColormap(d, 0/*DefaultScreenOfDisplay(d)*/);
  XAllocNamedColor(d, c_map, color, &c1, &c2);

  return c1.pixel;
}

int MakeSelectList(DfStr *s, DfFileList *list)
{
  DfFile *f;
  DfFile **selects;
  int n;
  int sel;

  f = list->files;
  sel = list->nSelects;
  if(sel == 0){
    Str_InitStr(s, f[list->nCur].pszName);
    Str_Shrink(s);
    return 1;
  }

  Str_Init(s);

  selects = bMalloc(sizeof(*selects) * sel);
  n = list->nFiles;
  while(n){
    if(f->nSelect){
      selects[f->nSelect - 1] = f;
      if(--sel == 0){
	/* all listed */
	break;
      }
    }
    f++;
    n--;
  }
  for(n = 0; n < list->nSelects; n++){
    Str_AddNUL(s, selects[n]->pszName);
  }
  Str_AddChar(s, '\0');
  Str_Shrink(s);
  
  bFree(selects);

  return list->nSelects;
}

char *FindLastChar(char *p, char c)
{
  char *f;

  f = NULL;
  while(*p){
    if(*p == c){
      f = p;
    }
    p++;
  }
  return f;
}

void InvalidateWindow(Widget w)
{
  XClearArea(XtDisplay(w), XtWindow(w), 0, 0, 0, 0, True);
}

void InvalidateRect(Widget w, XRectangle *rc)
{
  XClearArea(XtDisplay(w), XtWindow(w), rc->x, rc->y, rc->width, rc->height, True);
}


